/*       MONITOR             Ignacio  Merino  Lagunilla                       */

#include <unistd.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "global.hpp"

//#define 	DEBUG


monitor::monitor(void)
{
	activo = CIERTO;
	accesible = CIERTO;
	sig = ant = NULL;
}

monitor::~monitor(void)
{
#ifdef DEBUG
	printf("Destruyo el proceso monitor %d\n" , sock);
	if (fichero == CIERTO) fprintf(accion , "Destruyo el proceso monitor %d\n" , sock);
#endif
	if(sig != NULL) delete(sig);
}

int monitor::crear_primero(void)
{
 struct sockaddr_in dir;

	fin_m = this;
	if( (sock = socket(AF_INET , SOCK_STREAM , 0)) < 0)
	{
		printf("\aERROR: no se puede crear el socket principal del proceso monitor.\n");
		if (fichero == CIERTO) fprintf(accion , "ERROR: no se puede crear el socket principal del proceso monitor.\n");
		return(-1);
	}
#ifdef DEBUG
	printf("Creo el socket principal del proceso monitor\n");
	if (fichero == CIERTO) fprintf(accion , "Creo el socket principal del proceso monitor\n");
#endif
	bzero((char *) &dir , sizeof(struct sockaddr_in));
	dir.sin_family = AF_INET;
	dir.sin_addr.s_addr = htonl(INADDR_ANY);
	dir.sin_port = htons(puerto_m);
	if(bind(sock , (struct sockaddr*) &dir , sizeof(dir)) < 0)
	{
		printf("\aERROR: no se puede asociar la direccion al socket principal del proceso monitor.\n");
        	if (fichero == CIERTO) fprintf(accion , "ERROR: no se puede asociar la direccion al socket principal del proceso monitor.\n");
		close(sock);
		return(-2);
	}
#ifdef DEBUG
	printf("Asocio el socket principal del proceso monitor a su direccion\n");
	if (fichero == CIERTO) fprintf(accion , "Asocio el socket principal del proceso monitor a su direccion\n");
#endif
	if(listen(sock , 5))
	{
		printf("\aERROR: No se puede escuchar por el socket principal del proceso monitor.\n");
		if (fichero == CIERTO) fprintf(accion , "ERROR: No se puede escuchar por el socket principal del proceso monitor.\n");
		close(sock);
		return(-3);
	}
#ifdef DEBUG
	printf("Escucho por el socket principal del proceso monitor\n");
	if (fichero == CIERTO) fprintf(accion , "Escucho por el socket principal del proceso monitor\n");
#endif
	return(0);
}

int monitor::crear(void)
{
 struct sockaddr_in dir;
 int longitud = sizeof(struct sockaddr_in) , puerto;

	if( (sock = socket(AF_INET , SOCK_STREAM , 0)) < 0)
	{
		printf("\aERROR: No se puede crear el socket del proceso monitor.\n");
		if (fichero == CIERTO) fprintf(accion , "ERROR: No se puede crear el socket del proceso monitor.\n");
		return(-1);
	}
#ifdef DEBUG
	printf("Creo un socket para un proceso monitor\n");
	if (fichero == CIERTO) fprintf(accion , "Creo un socket para un proceso monitor\n");
#endif
	bzero((char *) &dir , longitud);
	dir.sin_family = AF_INET;
	dir.sin_addr.s_addr = htonl(INADDR_ANY);
	dir.sin_port = htons(0);
	if(bind(sock , (struct sockaddr *) &dir , sizeof(dir)) < 0)
	{
		printf("\aERROR: no se puede asociar la direccion al socket del proceso monitor.\n");
        	if (fichero == CIERTO) fprintf(accion , "ERROR: no se puede asociar la direccion al socket del proceso monitor.\n");
		close(sock);
		return(-2);
	}
#ifdef DEBUG
	printf("Asocio el socket del proceso monitor a su direccion\n");
	if (fichero == CIERTO) fprintf(accion , "Asocio el socket del proceso monitor a su direccion\n");
#endif
	if(listen(sock , 1))
	{
		printf("\aERROR: No se puede escuchar por el socket del proceso monitor.\n");
		if (fichero == CIERTO) fprintf(accion , "ERROR: No se puede escuchar por el socket del proceso monitor.\n");
		close(sock);
		return(-3);
	}
#ifdef DEBUG
	printf("Escucho por el socket del proceso monitor\n");
	if (fichero == CIERTO) fprintf(accion , "Escucho por el socket del proceso monitor\n");
#endif
	if(getsockname(sock , (struct sockaddr *) &dir , &longitud) < 0)
	{
		printf("\aERROR: no se puede acceder a los datos de socket.\n");
		if (fichero == CIERTO) fprintf(accion , "ERROR: no se puede acceder a los datos de socket.\n");
        	close(sock);
		return(-4);
	}
	puerto = htons(dir.sin_port);
#ifdef DEBUG
	printf("Obtengo el puerto %d para el proceso monitor\n" , puerto);
	if (fichero == CIERTO) fprintf(accion , "Obtengo el puerto %d para el proceso monitor\n" , puerto);
#endif
	ant = fin_m;
	fin_m->sig = this;
	fin_m = this;
#ifdef DEBUG
	printf("Enlazo el proceso monitor a la lista de monitores\n");
	if (fichero == CIERTO) fprintf(accion , "Enlazo el proceso monitor a la lista de monitores\n");
#endif
	return(puerto);
}

void monitor::destruir(int cerrar)
{
 monitor *fin;

#ifdef DEBUG
	printf("Destruyo el proceso monitor %d\n" , sock);
	if (fichero == CIERTO) fprintf(accion , "Destruyo el proceso monitor %d\n" , sock);
#endif
	if(cerrar != 0) shutdown(sock , 2);
	if(ant != NULL) ant->sig = sig;
	if(sig != NULL) { sig->ant = ant; sig = NULL; }
	if(fin_m == this) fin_m = ant;
	fin = this; delete(fin);
}

int monitor::aceptar_nuevos(void)
{
 int s , longitud , puerto , acceso = CIERTO;
 struct sockaddr_in cliente;
 fd_set test;
 monitor *ns;
 struct timeval espera;
 struct linger opcion;

 	FD_ZERO(&test);
	FD_SET(sock , &test);
	espera.tv_sec = 0;
	espera.tv_usec = 0;
	select(sock+1 , SOCKET &test , NULL , NULL , &espera);
	if(FD_ISSET(sock , &test))
	{
#ifdef DEBUG
		printf("Peticion de conexion de un nuevo proceso monitor\n");
		if (fichero == CIERTO) fprintf(accion , "Peticion de conexion de un nuevo proceso monitor\n");
#endif
		longitud = sizeof(cliente);
		if((s = accept(sock , (struct sockaddr *) &cliente , &longitud)) < 0)
		{
			printf("\aERROR: no se puede aceptar la conexion del proceso monitor.\n");
			if (fichero == CIERTO) fprintf(accion , "ERROR: no se puede aceptar la conexion del proceso monitor.\n");
			return(-1);
		}
		if ( (ns = new monitor) == NULL)
		{
			printf("\aERROR: no hay memoria suficiente para el proceso monitor.\n");
			if (fichero == CIERTO) fprintf(accion , "ERROR: no hay memoria suficiente para el proceso monitor.\n");
			close(s);
			return(-2);
		}
		if( (puerto = ns->crear()) < 0)
		{
			printf("\aERROR: No se puede crear el socket del proceso monitor.\n");
			if (fichero == CIERTO) fprintf(accion , "ERROR: No se puede crear el socket del proceso monitor.\n");
			close(s);
			return(-3);
		}
#ifdef DEBUG
		printf("Creo un nuevo socket para un proceso monitor con el puerto %d\n" , puerto);
		if (fichero == CIERTO) fprintf(accion , "Creo un nuevo socket para un proceso monitor con el puerto %d\n" , puerto);
#endif
		sprintf(msg_sale , "MNP%d" , puerto);
		if(enviar_msg(s , &acceso) < 0)
		{
			printf("\aERROR: No se puede enviar mensajes al proceso monitor.\n");
			if (fichero == CIERTO) fprintf(accion , "ERROR: No se puede enviar mensajes al proceso monitor.\n");
			ns->destruir(0);
			close(s);
			return(-4);
		}
#ifdef DEBUG
		printf("Envio al proceso monitor '%s'\n" , msg_sale);
		if (fichero == CIERTO) fprintf(accion , "Envio al proceso monitor '%s'\n" , msg_sale);
#endif
		FD_ZERO(&test);
		FD_SET(s , &test);
		espera.tv_sec = 3;
		espera.tv_usec = 0;
		select(s+1 , SOCKET &test , NULL , NULL , &espera);
		if(FD_ISSET(s , &test))
		{
			if(recibir_msg(s , &acceso) < 0)
			{
				printf("\aERROR: No se puede recibir mensajes del proceso monitor.\n");
				if (fichero == CIERTO) fprintf(accion , "ERROR: No se puede recibir mensajes del proceso monitor.\n");
				ns->destruir(0);
				close(s);
				return(-5);
			}
		}
		else
		{
			printf("\aERROR: No hay contestacion del proceso monitor.\n");
			if (fichero == CIERTO) fprintf(accion , "ERROR: No hay contestacion del proceso monitor.\n");
			ns->destruir(0);
			close(s);
			return(-6);
		}
#ifdef DEBUG
		printf("Recibo del proceso monitor '%s'\n" , msg_entra);
		if (fichero == CIERTO) fprintf(accion , "Recibo del proceso monitor '%s'\n" , msg_entra);
#endif
		if( strcmp("MAP" , msg_entra) != 0)
		{
			printf("\aERROR: Falla la confirmacion del proceso monitor.\n");
			if (fichero == CIERTO) fprintf(accion , "ERROR: Falla la confirmacion del proceso monitor.\n");
	        	ns->destruir(0);
			close(s);
			return(-7);
		}
		close(s);
		s = accept(ns->di_sock() , (struct sockaddr *) &cliente , &longitud);
		opcion.l_linger = 5;
		opcion.l_onoff = 1;
		setsockopt(s , SOL_SOCKET , SO_LINGER , &opcion , sizeof(opcion));
#ifdef DEBUG
		printf("Socket %d de cierre diferido\n" , s);
		if (fichero == CIERTO) fprintf(accion , "Socket %d de cierre diferido\n" , s);
#endif
		close(ns->di_sock());
		ns->fijar_sock(s);
#ifdef DEBUG
		printf("Acepto un nuevo proceso monitor en el socket %d\n" , s);
		if (fichero == CIERTO) fprintf(accion , "Acepto un nuevo proceso monitor en el socket %d\n" , s);
#endif
		sprintf(msg_sale , "MDM%d %d" , planox , planoy);
		if(enviar_msg(ns->sock , &acceso) < 0)
		{
			printf("\aERROR: No se puede enviar mensajes al proceso monitor.\n");
			if (fichero == CIERTO) fprintf(accion , "ERROR: No se puede enviar mensajes al proceso monitor.\n");
			ns->destruir(1);
			return(-8);
		}
		return(1);
	}
	return(0);
}

void monitor::responder_peticiones(void)
{
 char codigo[4] = { '\0' , '\0' , '\0' , '\0' };
 fd_set test;
 monitor *aux = this , *fin;
 struct timeval espera;

	while(aux != NULL)
	{
#ifdef DEBUG
		printf("Comprobando el proceso monitor %d\n" , aux->sock);
		if (fichero == CIERTO) fprintf(accion , "Comprobando el proceso monitor %d\n" , aux->sock);
#endif
		FD_ZERO(&test);
		FD_SET(aux->sock , &test);
		espera.tv_sec = 0;
		espera.tv_usec = 0;
		select(aux->sock + 1 , SOCKET &test , NULL , NULL , &espera);
		if(FD_ISSET(aux->sock , &test))
		{
			recibir_msg(aux->sock , &(aux->accesible));
#ifdef DEBUG
			printf("Recibo del proceso monitor %d: '%s'" , aux->sock , msg_entra);
			if (fichero == CIERTO) fprintf(accion , "Recibo del proceso monitor %d: '%s'" , aux->sock , msg_entra);
#endif
			strncpy(codigo , msg_entra , 3);
			if( strcmp(codigo , "MEF") == 0)
			{
				figuras->enviar_figuras(aux->sock , &(aux->accesible));
				sprintf(msg_sale , "MUF");
				enviar_msg(aux->sock , &(aux->accesible));
			}
			else if( strcmp(codigo , "MER") == 0)
			{
				ini_r->di_sig()->enviar_robots(aux->sock);
				sprintf(msg_sale , "MUR");
				enviar_msg(aux->sock , &(aux->accesible));
			}
			else if( strcmp(codigo , "MHP") == 0)
			{
				aux->activo = FALSO;
			}
			else if( strcmp(codigo , "MHC") == 0)
			{
				aux->activo = CIERTO;
			}
			else if( strcmp(codigo , "MHT") == 0)
			{
				strcpy(msg_sale , "MPT");
				enviar_msg(aux->sock , &(aux->accesible));
				printf("El proceso monitor %d quiere terminar\n" , aux->sock);
				if (fichero == CIERTO) fprintf(accion , "El proceso monitor %d quiere terminar\n" , aux->sock);
				fin = aux; aux = aux->ant; fin->destruir(1);
				ventana->redibujar(1);
			}
		}
		else
		{
#ifdef DEBUG
			printf("No hay ninguna peticion del proceso monitor %d\n" , aux->sock);
			if (fichero == CIERTO) fprintf(accion , "No hay ninguna peticion del proceso monitor %d\n" , aux->sock);
#endif
		}
#ifdef DEBUG
		printf("Proceso monitor %d, accesible = %d\n" , aux->sock , aux->accesible);
		if(fichero == CIERTO) fprintf(accion , "Proceso monitor %d, accesible = %d\n" , aux->sock , aux->accesible);
#endif
		aux=aux->sig;
	}
}

int monitor::escribir_lista(char *fich , int copia)
{
 FILE *mon;
 monitor *aux = this;
 int i = 0;
 static int num = 1;
 char nombre[50] , version[10];

	strcpy(nombre , fich);
	if(copia != 0)
	{
		sprintf(version , ".%d" , num);
		strcat(nombre , version);
		++num;
	}
	if ((mon = fopen(nombre , "wt")) == NULL)
	{
		printf("No puedo habrir el fichero '%s'\n" , nombre);
		if (fichero == CIERTO) fprintf(accion , "No puedo habrir el fichero '%s'\n" , nombre);
		return (-1);
	}
#ifdef DEBUG
	printf("Abro el fichero '%s' para escribir los procesos monitor\n" , nombre);
	if (fichero == CIERTO) fprintf(accion , "Abro el fichero '%s' para escribir los procesos monitor\n" , nombre);
#endif
	while(aux != NULL)
	{
		fprintf(mon , "Proceso monitor %d socket %d\n" , i , aux->sock);
		aux = aux->sig; ++i;
	}
#ifdef DEBUG
	printf("Escribo los %d procesos monitor del mundo\n" , i);
	if (fichero == CIERTO) fprintf(accion , "Escribo los %d procesos monitor del mundo\n" , i);
#endif
	fclose(mon);
	return(0);
}

void monitor::terminar(void)
{
 monitor *aux = this;

	sprintf(msg_sale , "MFY");
 	while(aux != NULL)
	{
		enviar_msg(aux->sock , &(aux->accesible));
		aux = aux->sig;
#ifdef DEBUG
		printf("Mando al proceso monitor %d '%s'\n" , aux->sock , msg_sale);
		if (fichero == CIERTO) fprintf(accion , "Mando al proceso monitor %d '%s'\n" , aux->sock , msg_sale);
#endif
	}
}

void monitor::enviar_figura(void)
{
 monitor *aux = this;

	while(aux != NULL)
	{
		if(aux->activo == CIERTO) enviar_msg(aux->sock , &(aux->accesible));
		aux = aux->sig;
	}
}

char* monitor::di_datos(void)
{
 static char linea[100];

 	sprintf(linea , "Monitor     %s" , (activo == CIERTO) ? "Activo" : "Parado" );
	return(linea);
}

void monitor::parar_monitores(void)
{
 monitor *aux = this;

	strcpy(msg_sale , "MPY");
	while(aux != NULL)
	{
		enviar_msg(aux->sock , &(aux->accesible));
		aux->activo = FALSO;
		aux = aux->sig;
	}
}

void monitor::continuar_monitores(void)
{
 monitor *aux = this;

	strcpy(msg_sale , "MCY");
	while(aux != NULL)
	{
		enviar_msg(aux->sock , &(aux->accesible));
		aux->activo = CIERTO;
		aux=aux->sig;
	}
}
