/******************************************************************************\
*                                                                              *
*       MUNDO                                 (C) Ignacio  Merino  Lagunilla   *
*                                                                              *
*       22/febrero/1995                                                        *
\******************************************************************************/

#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include "global.hpp"

//#define DEBUG


/*   Variables globales                                                       */

robot 		*ini_r = NULL;
robot		*fin_r = NULL;
monitor		*ini_m = NULL;
monitor		*fin_m = NULL;
figura		*figuras = NULL;
ventana_info 	*ventana = NULL;
mapa_mundo 	*mapa = NULL;
int			puerto_r = 7000;
int			puerto_m = 6500;
int 			final = FALSO;
int			activo = CIERTO;
int 			planox = 500;
int			planoy = 500;
char			msg_entra[MAX_MSG];
char			msg_sale[MAX_MSG];
int			fallo = FALSO;
FILE			*accion;
int 			fichero = FALSO;
int			pasos_dibujo = 1;
int			dib = 0;
char			f_figcop[50] = "figuras.cop";
char			f_mapcop[50] = "mapa.cop";
char			f_robcop[50] = "robot.cop";
char			f_moncop[50] = "monitor.cop";
char 		f_accion[50] = "mundo.acc";
char 		f_fig[50] = "figuras.dat";

/*   Manejadores de seales                                                   */

void ignorar(int sig);
void fin(int sig);

/*   Funcion principal                                                        */

int main(int argc , char *argv[])
{
 float paso = 0;
 int aux , ayuda = FALSO;

	system("clear");
	printf("\nSimDAI  V 1.0  -  MUNDO        (c) Ignacio Merino Lagunilla  1995\n\n");

/*   Compruebo los parametros de llamada al programa                          */
/*      Si la opcion es valida realizo la accion corespondiente               */
/*      Si la opcion no es valida muestro un mensaje indicandolo              */

	for(aux = 1 ; aux < argc ; aux++)
	{
		if(argv[aux][0] == '-')
		{
			switch(argv[aux][1])
			{
				case 'e': fichero = CIERTO;
				        if(argv[aux][2] != '\0') strcpy(f_accion , &argv[aux][2]);break;
				case 'f': strcpy(f_fig , &argv[aux][2]); break;
				case 'r': puerto_r = atoi(&argv[aux][2]); break;
				case 'm': puerto_m = atoi(&argv[aux][2]); break;
				case 'p': pasos_dibujo = atoi(&argv[aux][2]); break;
				case 'h':
				case '?': ayuda = CIERTO; break;
				default:  printf("%s: Opcion '%s' desconocida.\n" , argv[0] , argv[aux]);
					ayuda = CIERTO; break;
			}
		}
		else
		{
			printf("%s: Opcion '%s' desconocida.\n" , argv[0] , argv[aux]);
			ayuda = CIERTO;
		}
	}

/*   Si hay algun parametro erroneo o el usuario lo pide muestro como se      */
/*   llama al programa y todas las posibles opciones                          */
    
	if(ayuda == CIERTO)
	{
		printf("\nUso:\t%s [<opcion>...]\n\n" , argv[0]);
		printf("Las opciones posibles son:\n");
		printf("\t-h\t\tmuestra esta pantalla de ayuda.\n");
		printf("\t-?\t\tmuestra esta pantalla de ayuda.\n");
		printf("\t-e[<fichero>]\tescribe las acciones en el fichero dado o en el de defecto.\n");
		printf("\t-f<fichero>\tnombre del fichero con descripcion de las figuras.\n\n");
		printf("\t-r<puerto>\tpuerto para comunicarse con los procesos robot.\n");
		printf("\t-m<puerto>\tpuerto para comunicarse con los procesos monitor.\n");
		printf("\t-p<pasos>\tcada cuentos pasos se dibujan los robots.\n");
		return(1);
	}

/*   Instalo los distintos manejadores de seales                             */

	signal(SIGHUP , fin);
	signal(SIGINT , fin);
	signal(SIGQUIT , fin);
	signal(SIGILL , fin);
	signal(SIGFPE , ignorar);
	signal(SIGBUS , fin);
	signal(SIGSEGV , fin);
	signal(SIGPIPE , ignorar);
	signal(SIGTERM , fin);

/*   Si esta activada la opcion de escribir los mensajes en el fichero de     */
/*   acciones creo el fichero de acciones                                     */

	if(fichero == CIERTO)
	if( (accion = fopen(f_accion , "wt") ) == NULL)
	{
		printf("%s: No puedo abrir el fichero de acciones '%s'.\n\n" , argv[0] , f_accion);
		return(2);
	}
#ifdef DEBUG
	if(fichero == CIERTO) printf("Se escribiran las acciones en el fichero: '%s'\n" , f_accion);
	printf("El fichero de las figuras es: '%s'\n" , f_fig);
	printf("El puerto de los procesos robot es: %d\n" , puerto_r);
	printf("El puerto de los procesos monitor es: %d\n" , puerto_m);
	printf("Se dibujara cada %d pasos\n" , pasos_dibujo);
	if(fichero == CIERTO)
	{
		fprintf(accion , "Se escribiran las acciones en el fichero: '%s'\n" , f_accion);
		fprintf(accion , "El fichero de las figuras es: '%s'\n" , f_fig);
		fprintf(accion , "El puerto de los procesos robot es: %d\n" , puerto_r);
		fprintf(accion , "El puerto de los procesos monitor es: %d\n" , puerto_m);
		fprintf(accion , "Se dibujara cada %d pasos\n" , pasos_dibujo);
	}
#endif

/*   Reservo memoria para la ventana de informacion                           */

	if ( (ventana = new ventana_info) == NULL )
	{
		printf("%s: No hay memoria suficiente para la ventana de informacin.\n\n" , argv[0]);
		if(fichero == CIERTO)
		{
			fprintf(accion , "%s: No hay memoria suficiente para la ventana de informacion.\n" , argv[0]);
			fclose(accion);
		}
		return(3);
	}

/*   Reservo memoria para el mapa del mundo                                   */

	if ( (mapa = new mapa_mundo) == NULL )
	{
		printf("%s: No hay memoria suficiente para el mapa del mundo.\n\n" , argv[0]);
		if(fichero == CIERTO)
		{
			fprintf(accion , "%s: No hay memoria suficiente para el mapa del mundo.\n" , argv[0]);
			fclose(accion);
		}
		if(ventana != NULL) delete(ventana);
		return(4);
	}

/*   Creo la ventana de informacion                                           */

	ventana->crear();

/*   Creo las figuras del mundo desde el fichero de figuras                   */

	if( (figuras->crear(f_fig) < 0) )
	{
		printf("%s: No puedo crear las figuras del mundo.\n\n" , argv[0]);
		if(fichero == CIERTO)
		{
			fprintf(accion , "%s: No puedo crear las figuras del mundo.\n" , argv[0]);
			fclose(accion);
		}
		ventana->liberar();
		if(mapa != NULL) delete(mapa);
		if(ventana != NULL) delete(ventana);
		return(5);
	}

/*   Creo el mapa del mundo                                                   */

	if(mapa->crear() < 0)
	{
		printf("%s: No puedo crear el mapa del mundo.\n\n" , argv[0]);
		if(fichero == CIERTO)
		{
			fprintf(accion , "%s: No puedo crear el mapa del mundo.\n" , argv[0]);
			fclose(accion);
		}
		if(figuras != NULL) delete(figuras);
		ventana->liberar();
		if(mapa != NULL) delete(mapa);
		if(ventana != NULL) delete(ventana);
		return(6);
	}

/*   Dibujo las figuras en el mapa del mundo                                  */

	figuras->dibujar_figuras();

/*   Reservo memoria para el primer robot                                     */

	if ( (ini_r = new robot) == NULL )
	{
		printf("%s: No hay memoria suficiente para la lista de procesos robot.\n\n" , argv[0]);
		if(fichero == CIERTO)
		{
			fprintf(accion , "%s: No hay memoria suficiente para la lista de procesos robot.\n" , argv[0]);
			fclose(accion);
		}
		if(figuras != NULL) delete(figuras);
		ventana->liberar();
		if(mapa != NULL) delete(mapa);
		if(ventana != NULL) delete(ventana);
		return(7);
	}

/*   Creo el primer robot                                                     */

	if(ini_r->crear_primero() < 0)
	{
		printf("%s: No puedo crear la conexion con los procesos robot.\n\n" , argv[0]);
		if(fichero == CIERTO)
		{
			fprintf(accion , "%s: No puedo crear la conexion con los procesos robot.\n" , argv[0]);
			fclose(accion);
		}
		if(ini_r != NULL) delete(ini_r);
		if(figuras != NULL) delete(figuras);
		ventana->liberar();
		if(mapa != NULL) delete(mapa);
		if(ventana != NULL) delete(ventana);
		return(8);
	}

/*   Reservo memoria para el primer monitor                                   */

	if ( (ini_m = new monitor) == NULL )
	{
		printf("%s: No hay memoria suficiente para la lista de procesos monitor.\n\n" , argv[0]);
		if(fichero == CIERTO)
		{
			fprintf(accion , "%s: No hay memoria suficiente para la lista de procesos monitor.\n" , argv[0]);
			fclose(accion);
		}
		if(ini_r != NULL) delete(ini_r);
		if(figuras != NULL) delete(figuras);
		ventana->liberar();
		if(mapa != NULL) delete(mapa);
		if(ventana != NULL) delete(ventana);
		return(9);
	}

/*   Creo el primer monitor                                                   */

	if(ini_m->crear_primero() < 0)
	{
		printf("%s: No puedo crear la conexion con los procesos monitor.\n\n" , argv[0]);
		if(fichero == CIERTO)
		{
			fprintf(accion , "%s: No puedo crear la conexion con los procesos monitor.\n" , argv[0]);
			fclose(accion);
		}
		if(ini_m != NULL) delete(ini_m);
		if(ini_r != NULL) delete(ini_r);
		if(figuras != NULL) delete(figuras);
		ventana->liberar();
		if(mapa != NULL) delete(mapa);
		if(ventana != NULL) delete(ventana);
		return(10);
	}

/*   Bucle mientras el usuario no quiera finalizar                            */

	while (final == FALSO)
	{

/*   Comprobar si hay nuevos procesos robot                                   */

		aux=ini_r->aceptar_nuevos();
		switch(aux)
		{
			case 1:
				printf("Se acepta un nuevo proceso robot\n");
				if(fichero == CIERTO) fprintf(accion , "Se acepta un nuevo proceso robot\n");
				ventana->redibujar(1);
				break;
			case 0:
#ifdef DEBUG
				printf("No hay nuevos procesos robot\n");
				if(fichero == CIERTO) fprintf(accion , "No hay nuevos procesos robot\n");
#endif
				break;
			case -1:
			case -2:
			case -3:
			case -4:
			case -5:
				printf("\aERROR: Hay un error en la conexion con un nuevo proceso robot.\n");
				if(fichero == CIERTO) fprintf(accion , "ERROR: Hay un error en la conexion con un nuevo proceso robot.\n");
				break;
			case -6:
				printf("\aERROR: Hay un error en los datos del robot.\n");
				if(fichero == CIERTO) fprintf(accion , "ERROR: Hay un error en los datos del robot.\n");
				break;
			default:
				printf("\aERROR: Retorno anormal en la conexion con un nuevo proceso robot.\n");
				if(fichero == CIERTO) fprintf(accion , "ERROR: Retorno anormal en la conexion con un nuevo proceso robot.\n");
				break;
		}

/*   Responder a las peticiones de los procesos robot                         */

		ini_r->di_sig()->responder_peticiones();

/*   Comprobarn si hay nuevos procesos monitor                                */

		aux=ini_m->aceptar_nuevos();
		switch(aux)
		{
			case 1:
				printf("Se acepta un nuevo proceso monitor\n");
				if(fichero == CIERTO) fprintf(accion , "Se acepta un nuevo proceso monitor\n");
				ventana->redibujar(1);
				break;
			case 0:
#ifdef DEBUG
				printf("No hay nuevos procesos monitor\n");
				if(fichero  == CIERTO) fprintf(accion , "No hay nuevos procesos monitor\n");
#endif
				break;
			case -1:
			case -2:
			case -3:
			case -4:
			case -5:
				printf("\aERROR: Hay un error en la conexion con un nuevo proceso monitor.\n");
				if(fichero == CIERTO) fprintf(accion , "ERROR: Hay un error en la conexion con un nuevo proceso monitor.\n");
				break;
			default:
				printf("\aERROR: Retorno anormal en la conexion con un nuevo proceso monitor.\n");
				if(fichero == CIERTO) fprintf(accion , "ERROR: Retorno anormal en la conexion con un nuevo proceso monitor.\n");
				break;
		}

/*   Responder a las peticiones de los procesos monitor                       */

		ini_m->di_sig()->responder_peticiones();

/*   Responder a las peticiones de los procesos robot                         */

		ini_r->di_sig()->responder_peticiones();

/*   Muevo a todos los robots si estoy activo                                 */

		if(activo == CIERTO) ini_r->di_sig()->mover();

/*   Escribo la informacion en la ventana                                     */

		ventana->redibujar(0);

/*   Calculo los pasos que llevo y compruebo si toca dibujar en este paso.    */

		dib = (dib + 1) % pasos_dibujo;
#ifdef DEBUG
		printf("Realizo el paso %f\n" , ++paso);
		if(fichero == CIERTO) fprintf(accion , "Realizo el paso %f\n" , paso);
#endif

/*   Se comprueban y se realizan los eventos que se hayan producido           */

		ventana->comprobar_eventos();

/* compruebo si se ha producido algun fallo con alguna conexion               */

		if(fallo == CIERTO) comprobar_procesos();

	}
#ifdef DEBUG
	printf("Termino la ejecucion del programa, final = %d\n" , final);
	if(fichero == CIERTO) fprintf(accion , "Termino la ejecucion del programa, final = %d\n" , final);
#endif

/*   Se notifica a los procesos monitor que se finaliza                       */

	if(ini_m->di_sig() != NULL) ini_m->di_sig()->terminar();

/*   Se destruye la lista de los procesos monitor                             */

	if(ini_m != NULL) delete(ini_m);
#ifdef DEBUG
	printf("Destruyo la lista de los procesos monitor\n");
	if(fichero == CIERTO) fprintf(accion , "Destruyo la lista de los procesos monitor\n");
#endif

/*   Se notifica a los procesos robot que se finaliza                         */

	if(ini_r->di_sig() != NULL) ini_r->di_sig()->terminar();

/*   Se destruye la lista de los procesos robot                               */

	if(ini_r != NULL) delete(ini_r);
#ifdef DEBUG
	printf("Destruyo la lista de los procesos robot\n");
	if(fichero == CIERTO) fprintf(accion , "Destruyo la lista de los procesos robot\n");
#endif

/*   Desturyo la lista de figuras                                             */

	if(figuras != NULL) delete(figuras);
#ifdef DEBUG
	printf("Destruyo la lista de las figuras del mundo\n");
	if(fichero == CIERTO) fprintf(accion , "Destruyo la lista de las figuras del mundo\n");
#endif

/*   Se libera la ventana de informacion                                      */

	ventana->liberar();

/*   Se libera la memoria del mapa del mundo                                  */

	if(mapa != NULL) delete(mapa);
#ifdef DEBUG
	printf("Destruyo el mapa del mundo\n");
	if(fichero == CIERTO) fprintf(accion , "Destruyo el mapa del mundo\n");
#endif

/*   Se libera la memoria de la ventana de informacion                        */

	if(ventana != NULL) delete(ventana);
#ifdef DEBUG
	printf("Destruyo la ventana de informacion del mundo\n");
	if(fichero == CIERTO) fprintf(accion , "Destruyo la ventana de informacion del mundo\n");
#endif

/*   Si esta activa la opcion de generar el fichero de acciones se cierra     */

	if(fichero == CIERTO) fclose(accion);

/*   Se termina la ejecucion del programa                                     */

	return(0);
}


/*   Este manajador de seales captura las seales que queremos ignorar, y    */
/*   simplemente cuando el sistema le avisa de una seal muestra un mensaje   */
/*   indicando que dicha seal es ignorada                                    */

void ignorar(int sig)
{
	signal(SIGPIPE , ignorar);
	signal(SIGFPE , ignorar);
	printf("\nIgnoro la signal: %d\n" , sig);
	if(fichero == CIERTO) fprintf(accion , "\nIgnoro la signal %d\n" , sig);
}


/*   Este manejador captura las seales que nos avisan que se ha producido un */
/*   error grave. Cuando el sistema le avisa de una de estas seales, termina */
/*   con todo lo que el programa este aciendo en ese momento y finaliza la    */
/*   ejecucion del programa                                                   */

void fin(int sig)
{
	signal(SIGQUIT , SIG_IGN);
	signal(SIGINT , SIG_IGN);
	signal(SIGHUP , SIG_IGN);
	signal(SIGILL , SIG_IGN);
	signal(SIGBUS , SIG_IGN);
	signal(SIGSEGV , SIG_IGN);
	signal(SIGTERM , SIG_IGN);
	printf("\nCapturada la interrupcion: %d\nTerminando...\n" , sig);
	if(fichero == CIERTO) fprintf(accion , "\nCapturada la interrupcion: %d\nTerminando...\n" , sig);
	if(ini_m != NULL && ini_m->di_sig() != NULL) ini_m->di_sig()->terminar();
	if(ini_m != NULL) delete(ini_m);
#ifdef DEBUG
	printf("Destruyo la lista de los procesos monitor\n");
	if(fichero == CIERTO) fprintf(accion , "Destruyo la lista de los procesos monitor\n");
#endif
	if(ini_r != NULL && ini_r->di_sig() != NULL) ini_r->di_sig()->terminar();
	if(ini_r != NULL) delete(ini_r);
#ifdef DEBUG
	printf("Destruyo la lista de los procesos robot\n");
	if(fichero == CIERTO) fprintf(accion , "Destruyo la lista de los procesos robot\n");
#endif
	if(figuras != NULL) delete(figuras);
#ifdef DEBUG
	printf("Destruyo la lista de las figuras del mundo\n");
	if(fichero == CIERTO) fprintf(accion , "Destruyo la lista de las figuras del mundo\n");
#endif
	if(ventana != NULL) ventana->liberar();
	if(mapa != NULL) delete(mapa);
#ifdef DEBUG
	printf("Destruyo el mapa del mundo\n");
	if(fichero == CIERTO) fprintf(accion , "Destruyo el mapa del mundo\n");
#endif
	if(ventana != NULL) delete(ventana);
#ifdef DEBUG
	printf("Destruyo la ventana de informacion del mundo\n");
	if(fichero == CIERTO) fprintf(accion , "Destruyo la ventana de informacion del mundo\n");
#endif
	if(fichero == CIERTO) fclose(accion);
	exit(20);
}
