/******************************************************************************\
*                                                                              *
*       SENSORES            Ignacio  Merino  Lagunilla                         *
*                                                                              *
*       Funciones para definir los distintos sensores de los robots.           *
*                                                                              *
*       22/febrero/1995                                                        *
\******************************************************************************/

#ifndef _SENSORES_
#define _SENSORES_

#include <stdio.h>
#include "coord.hpp"

class robot;

class sensor{
	protected:
		punto_pol pos;
		float *salida;
		sensor *sig;
		sensor *ant;
	public:
		inline sensor(void) { sig = ant = NULL; salida = NULL; }
		virtual ~sensor(void) { }
		inline sensor* di_sig(void) { return(sig); }
		inline void fijar_sig(sensor *s) { sig = s; }
		inline sensor* di_ant(void) { return(ant); }
		inline void fijar_ant(sensor *a) { ant = a; }
		void explorar_sensores(robot *rob);
		virtual void explorar(robot *rob , float *salida) = 0;
		virtual void crear(sensor *&ini , sensor *&fin , char *datos) = 0;
};

class laser:public sensor{
	private:
		float angulo;
		int rango;
	public:
		inline laser(void) { salida = new float[2]; }
		inline ~laser(void) { if(sig != NULL) delete(sig); if(salida != NULL) delete(salida); }
		void crear(sensor *&ini , sensor *&fin , char *datos);
		void explorar(robot *rob , float *salida);
};

class sonar:public sensor{
	private:
		float angulo;
		int rango;
		float ang_sec;					// angulo del sector
		float intervalo;				// intervalo exploracion laser
		char tipo;					// 1=minimo 2=media 3=maximo
	public:
		inline sonar(void) { salida = new float[2]; }
		inline ~sonar(void) { if(sig != NULL) delete(sig); if(salida != NULL) delete(salida); }
		void crear(sensor *&ini , sensor *&fin , char *datos);
		void explorar(robot *rob , float *salida);
};

class GPS:public sensor{
	private:
	public:
		inline GPS(void) { salida = new float[3]; }
		inline ~GPS(void) { if(sig != NULL) delete(sig); if(salida != NULL) delete(salida); }
		void crear(sensor *&ini , sensor *&fin , char *datos);
		void explorar(robot *rob , float *salida);
};

#endif
