/******************************************************************************\
*                                                                              *
*       ROBOT                               (C) Ignacio  Merino  Lagunilla     *
*                                                                              *
*       22/febrero/1995                                                        *
\******************************************************************************/

#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include "global.hpp"

//#define		DEBUG

/*   Variables globales                                                       */

ventana_info	*ventana = NULL;
robot_inf		*ini_ri = NULL;
robot_inf		*fin_ri = NULL;
int			puerto_r = 7000;
char			servidor[50] = "SimDAI";
int			final = FALSO;
int			acabar = FALSO;
int			conectado = FALSO;
int 			fichero = FALSO;
char			msg_entra[MAX_MSG];
char			msg_sale[MAX_MSG];
FILE 		*accion;
robot		*rob = NULL;
char 		f_accion[50] = "robot.acc";
char 		f_rob[50] = "robot.dat";

/*   Manejadores de seales                                                   */

void ignorar(int sig);
void fin(int sig);

/*   Funcion principal                                                        */

int main(int argc , char *argv[])
{
 int aux , ayuda = FALSO;

	system("clear");
	printf("\nSimDAI  V 1.0  -  ROBOT        (c) Ignacio Merino Lagunilla  1995\n\n");

/*   Compruebo los parametros de llamada al programa                          */
/*      Si la opcion es valida realizo la accion corespondiente               */
/*      Si la opcion no es valida muestro un mensaje indicandolo              */

	for(aux = 1 ; aux < argc ; aux++)
	{
		if(argv[aux][0] == '-')
		{
			switch(argv[aux][1])
			{
				case 'e': fichero = CIERTO;
				        if(argv[aux][2] != '\0') strcpy(f_accion , &argv[aux][2]);break;
				case 'd': strcpy(f_rob , &argv[aux][2]); break;
				case 's': strcpy(servidor , &argv[aux][2]); break;
				case 'r': puerto_r = atoi(&argv[aux][2]); break;
				case 'h':
				case '?': ayuda = CIERTO; break;
				default:  printf("%s: Opcion '%s' desconocida.\n" , argv[0] , argv[aux]);
					ayuda = CIERTO; break;
			}
		}
		else
		{
			printf("%s: Opcion '%s' desconocida.\n" , argv[0] , argv[aux]);
			ayuda = CIERTO; break;
		}
	}

/*   Si hay algun parametro erroneo o el usuario lo pide muestro como se      */
/*   llama al programa y todas las posibles opciones                          */
    
	if(ayuda == CIERTO)
	{
		printf("\nUso:\t%s [<opcion>...]\n" , argv[0]);
		printf("Las opciones posibles son:\n");
		printf("\t-h\t\tmuestra esta pantalla de ayuda.\n");
		printf("\t-?\t\tmuestra esta pantalla de ayuda.\n");
		printf("\t-d<fichero>\tnombre del fichero con la descripcion del robot.\n");
		printf("\t-e[<fichero>]\tescribe las acciones en el fichero dado o en el de defecto.\n");
		printf("\t-s<servidor>\tnombre del servidor donde conectarse con el mundo.\n");
		printf("\t-r<puerto>\tpuerto para comunicarse con el mundo.\n");
		return(1);
	}

/*   Instalo los distintos manejadores de seales                             */

	signal(SIGHUP , fin);
	signal(SIGINT , fin);
	signal(SIGQUIT , fin);
	signal(SIGILL , fin);
	signal(SIGFPE , ignorar);
	signal(SIGBUS , fin);
	signal(SIGSEGV , fin);
	signal(SIGPIPE , fin);
	signal(SIGTERM , fin);

/*   Si esta activada la opcion de escribir los mensajes en el fichero de     */
/*   acciones creo el fichero de acciones                                     */

	if(fichero == CIERTO)
	if ((accion = fopen(f_accion , "wt")) == NULL)
	{
		printf("%s: No puedo abrir el fichero de acciones '%s'.\n\n" , argv[0] , f_accion);
		return(2);
	}
#ifdef DEBUG
	if(fichero == CIERTO) printf("Se escribiran las acciones en el fichero: '%s'\n" , f_accion);
	printf("El fichero de los datos del robot es: '%s'\n" , f_rob);
	printf("El puerto del proceso robot es: %d\n" , puerto_r);
	printf("El nombre del servidor es: '%s'\n" , servidor);
	if(fichero == CIERTO)
	{
		fprintf(accion , "Se escribiran las acciones en el fichero: '%s'\n" , f_accion);
		fprintf(accion , "El fichero de los datos del robot es: '%s'\n" , f_rob);
		fprintf(accion , "El puerto del proceso robot es: %d\n" , puerto_r);
		fprintf(accion , "El nombre del servidor es: '%s'\n" , servidor);
	}
#endif

/*   Reservo memoria para la ventana de informacion                           */

	if ( (ventana = new ventana_info) == NULL )
	{
		printf("%s: No hay memoria suficiente para la ventana de informacin.\n\n" , argv[0]);
		if(fichero == CIERTO)
		{
			fprintf(accion , "%s: No hay memoria suficiente para la ventana de informacion.\n" , argv[0]);
			fclose(accion);
		}
		return(3);
	}

/*   Reservo memoria para el robot                                            */

	if ( (rob = new robot) == NULL )
	{
		printf("%s: No hay memoria suficiente para el robot.\n\n" , argv[0]);
		if(fichero == CIERTO)
		{
			fprintf(accion , "%s: No hay memoria suficiente para el robot.\n" , argv[0]);
			fclose(accion);
		}
		if(ventana != NULL) delete(ventana);
		return(4);
	}

/*   Creo el robot leyendo sus datos de su fichero de descripcion             */

	if(rob->crear(f_rob) < 0)
	{
		printf("%s: No puedo crear el robot.\n\n" , argv[0]);
		if(fichero == CIERTO)
		{
			fprintf(accion , "%s: No puedo crear el robot.\n" , argv[0]);
			fclose(accion);
		}
		if(rob != NULL) delete(rob);
		if(ventana != NULL) delete(ventana);
		return(5);
	}

/*   Creo la ventana de informacion donde se visualizan los datos del robot   */

	ventana->crear();

/*   Bucle mientras el usuario no quiera finalizar                            */

	while(final == FALSO)
	{
		acabar = FALSO;

/*   Mientras no este conectado con el mundo solo compruebo eventos           */

		while (conectado == FALSO && final == FALSO) ventana->comprobar_eventos();

/*   Bucle en el que el robot realiza sus acciones y se comprueban eventos    */

		while (acabar == FALSO)
		{

			if(rob->di_activo() == CIERTO)
			{

/*   El robot lee todos sus sensores                                          */

				rob->leer_sensores();

/*   Decide que velocidad ha de poner en las dos ruedas que tiene             */

				rob->decidir();

/*   Se escribe la informacion del robot en la ventana                        */

				ventana->redibujar();

			}

/*   Se comprueba si hay algun mensaje                                        */

			rob->recibir_msg(rob->di_sock() , "***" , 1);

/*   Se comprueban y se realizan los eventos que se hayan producido           */

			ventana->comprobar_eventos();

		}
	}

/*   Se libera la ventana de informacion                                      */

	ventana->liberar();

/*   Se libera la memoria del robot                                           */

	if(rob != NULL) delete(rob);

/*   Se libera la memoria de la ventana de informacion                        */

	if(ventana != NULL) delete(ventana);

/*   Se libera la memoria de la lista de robots que se conocen                */

	if(ini_ri != NULL) delete(ini_ri);

/*   Si esta activa la opcion de generar el fichero de acciones se cierra     */

	if(fichero == CIERTO) fclose(accion);

/*   Se termina la ejecucion del programa                                     */

	return(0);
}


/*   Este manajador de seales captura las seales que queremos ignorar, y    */
/*   simplemente cuando el sistema le avisa de una seal muestra un mensaje   */
/*   indicando que dicha seal es ignorada                                    */

void ignorar(int sig)
{
	signal(SIGFPE , ignorar);
	printf("\nIgnoro la signal: %d\n" , sig);
	if(fichero == CIERTO) fprintf(accion , "\nIgnoro la signal %d\n" , sig);
}


/*   Este manejador captura las seales que nos avisan que se ha producido un */
/*   error grave. Cuando el sistema le avisa de una de estas seales, termina */
/*   con todo lo que el programa este aciendo en ese momento y finaliza la    */
/*   ejecucion del programa                                                   */

void fin(int sig)
{
	signal(SIGPIPE , SIG_IGN);
	signal(SIGTERM , SIG_IGN);
	signal(SIGHUP , SIG_IGN);
	signal(SIGQUIT , SIG_IGN);
	signal(SIGILL , SIG_IGN);
	signal(SIGINT , SIG_IGN);
	signal(SIGBUS , SIG_IGN);
	signal(SIGSEGV , SIG_IGN);
	printf("\nCapturada la interrupcion: %d\nTerminando...\n" , sig);
	if(fichero == CIERTO) fprintf(accion , "\nCapturada la interrupcion: %d\nTerminando...\n" , sig);
	if(conectado == CIERTO) rob->desconectar();
	if(ventana != NULL) ventana->liberar();
	if(rob != NULL) delete(rob);
	if(ventana != NULL) delete(ventana);
	if(ini_ri != NULL) delete(ini_ri);
	if(fichero == CIERTO) fclose(accion);
	exit(20);
}
